"""Diagnostics client."""

from enum import Enum, unique
from typing import Dict, cast

from fred_api.diagnostics.service_diagnostics_grpc_pb2 import StatusReply
from fred_api.diagnostics.service_diagnostics_grpc_pb2_grpc import DiagnosticsStub
from fred_types import BaseModel
from frgal import GrpcDecoder
from frgal.aio import AsyncGrpcClient
from google.protobuf.empty_pb2 import Empty

from ._utils import catch_errors


@unique
class Status(str, Enum):
    """Service status code."""

    UNKNOWN = "unknown"
    OK = "ok"
    WARNING = "warning"
    ERROR = "error"


class About(BaseModel):
    """Information about server and API versions."""

    server_version: str
    api_versions: Dict[str, str]


class Service(BaseModel):
    """Service status information."""

    status: Status
    note: str = ""
    extras: Dict[str, str] = {}


class DiagnosticsDecoder(GrpcDecoder):
    """Diagnostics service decoder."""

    decode_unset_messages = False
    decode_empty_string_none = False

    STATUS_CODE_ENUM = {
        0: Status.UNKNOWN,
        1: Status.OK,
        2: Status.WARNING,
        3: Status.ERROR,
    }

    def __init__(self) -> None:
        super().__init__()
        self.set_decoder(StatusReply.Data.Service, self._decode_service)

    def _decode_service(self, value: StatusReply.Data.Service) -> Service:
        """Decode service status information."""
        return Service(
            status=self.STATUS_CODE_ENUM[value.status],
            note=self.decode(value.note),
            extras=self.decode(value.extras),
        )


class DiagnosticsClient(AsyncGrpcClient):
    """Diagnostics service gRPC client."""

    service = "Diagnostics"
    decoder_cls = DiagnosticsDecoder
    stub_cls = DiagnosticsStub

    @catch_errors
    async def about(self) -> About:
        """Return information about server and API versions.

        Raises:
            DiagnosticsNotImplemented: If diagnostics service is not implemented
                on the server.
            ServiceUnavailable: If connection to the service failed.
            DiagnosticsError: If there has been any other problem with the response.
        """
        result = await self.call("about", Empty())
        return About(**result or {})

    @catch_errors
    async def status(self) -> Dict[str, Service]:
        """Return service status information for all available services.

        Raises:
            DiagnosticsNotImplemented: If diagnostics service is not implemented
                on the server.
            ServiceUnavailable: If connection to the service failed.
            DiagnosticsError: If there has been any other problem with the response.
        """
        result = await self.call("status", Empty())
        return cast(Dict[str, Service], result or {})
